local mod = Vigilant:NewModule("Vigilance", nil, "AceEvent-3.0", "AceConsole-3.0", "AceTimer-3.0", "LibSink-2.0")
local Media = LibStub("LibSharedMedia-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Vigilant")
local self = mod
local db

local defaults = {
	profile = {
		vigilance = {
			y = 350,
			
			width   = 250,
			height  = 16,
			alpha   = 1,
			scale   = 100.0,
			texture = "BantoBar",
			icon = true,
			iconPosition = 1,
			
			border    = "RothSquare",
			backColor = { 0, 0, 0, 1 },
			barColor = { 199/255, 156/255, 110/255, 0.8 },
			barColorByClass = true,
			
			textShow     = true,
			textPosition = 2,
			textColor    = { 1, 1, 1, 1 },
			textFont     = "Emblem",
			textSize     = 14,
			
			sink = {},
		}
	}
}

-- Frames
local locked = true
local vgBar, vgTimeText, vgParent

-- Localized functions
local GetTime = _G.GetTime
local UnitGUID = _G.UnitGUID
local playerName = nil

-- Settings/infos
local resetValues = true
local spellInfo = GetSpellInfo(50720) --Vigilance

local function OnUpdate()
	if self.running then
		local currentTime = GetTime()
		local startTime = self.startTime
		local endTime = self.endTime
	
		local remainingTime = endTime - currentTime
		remainingTime = ((remainingTime > 0) and remainingTime or 0) -- Hackity hack hack
	
		if remainingTime == 0 then
			self.running = false
		else
			local perc = remainingTime / (30 * 60)
			vgBar:SetValue(perc)
			if remainingTime > 60 then
				vgTimeText:SetText(("%dm"):format(remainingTime / 60) .. " - " .. mod.VigilanceTarget.name)
			else			
				vgTimeText:SetText(("%.1f"):format(remainingTime) .. " - " .. mod.VigilanceTarget.name)
			end
		end
	else
		if vgParent:IsVisible() then
			self:CheckVisibility(true)
		end
	end
end
mod.OnUpdate = OnUpdate
local function OnShow()
	vgParent:SetScript("OnUpdate", OnUpdate)
end
local function OnHide()
	vgParent:SetScript("OnUpdate", nil)
end

function mod:OnInitialize()
	db = LibStub("AceDB-3.0"):New("VigilantDB", nil, "Default")
	self.db = db
	
	self.db:RegisterDefaults(defaults)
	
	self:SetEnabledState(false)
end

function mod:OnEnable()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("RAID_ROSTER_UPDATE")
	
	playerName = UnitName("player")
	
	self.locked = locked
	self.startTime = 0
	self.endTime = 0
	self.running = false
	self.VigilanceTarget = {
		name,
		class,
	}

	if not vgParent then
		vgParent = CreateFrame("Frame", nil, UIParent)
		vgParent:SetFrameStrata("LOW")
		vgParent:SetScript("OnShow", OnShow)
		vgParent:SetScript("OnHide", OnHide)
		vgParent:SetMovable(true)
		vgParent:RegisterForDrag("LeftButton")
		vgParent:SetClampedToScreen(true)
		
		vgBar = CreateFrame("StatusBar", nil, vgParent)
		vgBar:SetFrameStrata("MEDIUM")
		vgTimeText = vgBar:CreateFontString(nil, "OVERLAY")
		
		vgParent:Hide()
	end
	self:ApplySettings()
	self:SetSinkStorage(db.profile.vigilance.sink)
end

function mod:OnDisable()
	self.running = false
	self.locked = true
	vgParent:Hide()
	vgParent = nil
	
	self:UnregisterAllEvents()
end

-- ---------------------
-- Frame methods
-- ---------------------

function mod:ApplySettings()
	if not self:IsEnabled() then return end
	
	if vgParent then
		local db = db.profile.vigilance
		local back = {}
		
		-- vgParent, to which all our stuff is anchored
		vgParent:ClearAllPoints()
		if not db.x then
			db.x = (UIParent:GetWidth() / 2 - (db.width * (db.scale/100.0)) / 2) / (db.scale/100.0)
		end
		vgParent:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", db.x, db.y)
		
		if db.border == "None" then
			vgParent:SetWidth(db.width)
			vgParent:SetHeight(db.height)
			
			back.bgFile = "Interface\\Tooltips\\UI-Tooltip-Background"
			back.tile = true
			back.tileSize = 16
			back.insets = { top = 0, right = 0, bottom = 0, left = 0 }
		else
			vgParent:SetWidth(db.width + 9)
			vgParent:SetHeight(db.height + 10)
			
			back.bgFile = "Interface\\Tooltips\\UI-Tooltip-Background"
			back.tile = true
			back.tileSize = 16
			back.edgeFile = Media:Fetch("border", db.border)
			back.edgeSize = 16
			back.insets = { top = 4, right = 4, bottom = 4, left = 4 }
		end
		vgParent:SetBackdrop(back)
		vgParent:SetBackdropColor(unpack(db.backColor))
		
		vgParent:SetAlpha(db.alpha)
		vgParent:SetScale(db.scale / 100.0)
		
		-- make us an icon!
		vgParent.icon = vgParent.icon or vgParent:CreateTexture(nil, "OVERLAY")
		vgParent.icon:SetPoint("RIGHT", vgParent, "LEFT", 0, 0)
		vgParent.icon:SetHeight(db.height + 5) --scaled 1.1
		vgParent.icon:SetWidth(db.height + 5)
		vgParent.icon:SetTexture([[Interface\ICONS\Ability_Warrior_Vigilance]])
		if db.icon then 
			vgParent.icon:Show() 
		else
			vgParent.icon:Hide()
		end
		
		-- vgBar, the actual timer bar
		vgBar:ClearAllPoints()
		vgBar:SetPoint("CENTER", vgParent, "CENTER")
		vgBar:SetWidth(db.width)
		vgBar:SetHeight(db.height)
		vgBar:SetStatusBarTexture(Media:Fetch("statusbar", db.texture))
		vgBar:SetMinMaxValues(0, 1)
		if not db.barColorByClass then
			vgBar:SetStatusBarColor(unpack(db.barColor))
		end	
		vgBar:Show()
		
		-- vgTimeText, countdown timer text
		vgTimeText:ClearAllPoints()
		if db.textPosition == 1 then -- LEFT
			vgTimeText:SetPoint("LEFT", vgParent, "LEFT", 5, 0)
			vgTimeText:SetJustifyH("LEFT")
		elseif db.textPosition == 2 then -- CENTER
			vgTimeText:SetPoint("CENTER", vgParent, "CENTER")
			vgTimeText:SetJustifyH("CENTER")
		elseif db.textPosition == 3 then -- RIGHT
			vgTimeText:SetPoint("RIGHT", vgParent, "RIGHT", -5, 0)
			vgTimeText:SetJustifyH("RIGHT")
		end
		vgTimeText:SetFont(Media:Fetch("font", db.textFont), db.textSize)
		vgTimeText:SetTextColor(unpack(db.textColor))
		vgTimeText:SetShadowColor(0, 0, 0, 1)
		vgTimeText:SetShadowOffset(0.8, -0.8)
		vgTimeText:SetNonSpaceWrap(false)
		if db.textShow then
			vgTimeText:Show()
		else
			vgTimeText:Hide()
		end
		
		self:CheckVisibility(true)
		
		-- If we're not already running a timer, set some sane default values
		-- These are used when the user unlocks the bar, they can get an idea of
		-- what it'll look like. But we don't want to change these values when the
		-- bar's running!
		if not self.running and not self.locked then
			resetValues = true

			vgBar:SetValue(0.30)
			vgTimeText:SetText(L["Vigilance"])
		end
	end
end

-- Checks whether or not to show our parent frame based on various conditions
-- Args: perform - true to call Hide or Show on the frame based on results
-- returns true if the frame should be shown
function mod:CheckVisibility(perform)
	local visible = false
	
	-- If we're running, we're visible. Simple as that.
	if self.running then
		visible = true
	else
		-- We're unlocked and want to give the user something to drag/test settings with
		if not self.locked then
			visible = true
		end
	end
	
	if perform then
		if visible then
			vgParent:Show()
		else
			vgParent:Hide()
		end
	end
	
	return visible
end

function mod:VigiScan()
	if not self.VigilanceTarget.name then return end
	name, _, _, _, _, _, _, unitCaster = UnitAura(self.VigilanceTarget.name, "Vigilance")
	if name then
		if not unitCaster then return end --target out of range, blindly assuming its ours
		if unitCaster ~= "player" then --someone overwrote ours!
			self:CancelBar()
			return
		end
		--do nuffink! vigilance still active on our target
	else
		self:CancelBar()
	end
end

function mod:RAID_ROSTER_UPDATE()
	if not self.VigilanceTarget.name then return end
	if (GetNumRaidMembers() == 0) or (not UnitInRaid(self.VigilanceTarget.name)) then
		self:CancelBar()
		return
	end
end

function mod:StartBar(target, length)
	--set aside the name and class of our target
	self.VigilanceTarget.name = target
	self.VigilanceTarget.class = select(2, UnitClass(target))

	--establish tracking bar values and get it running
	if db.profile.vigilance.barColorByClass then
		local r = RAID_CLASS_COLORS[self.VigilanceTarget.class].r
		local g = RAID_CLASS_COLORS[self.VigilanceTarget.class].g
		local b = RAID_CLASS_COLORS[self.VigilanceTarget.class].b
		
		vgBar:SetStatusBarColor(r, g, b, .8)
	end
	self.startTime = GetTime()
	self.endTime = self.startTime + length
	vgBar:SetValue(1)
	self.running = true	
	vgParent:Show()

	--schedule to check vigilance on our target for out of range s	cenarios
	self.vigiscantimer = self:ScheduleRepeatingTimer("VigiScan", 1)

	--if we're in a raid, alert of our choosing declaring vigilance on our target
	if GetNumRaidMembers() > 0 --[[and db.profile.vigilance.alertraid]] then
		self:Pour("Vigilance on " .. self.VigilanceTarget.name)
	end
end

function mod:CancelBar()
	self.running = false
	vgParent:Hide()
	
	RaidNotice_AddMessage(RaidWarningFrame, "Vigilance expired on ".. self.VigilanceTarget.name, RAID_CLASS_COLORS[self.VigilanceTarget.class])
	self.VigilanceTarget.name, self.VigilanceTarget.class = nil, nil
	self:CancelTimer(self.vigiscantimer)
end

-- ---------------------
-- Debugging
-- ---------------------

function mod:TestBar()
	if not self:IsEnabled() then return end
	
	--vgBar:SetValue(1)
	self:StartBar("TestDummy", 60 * 2)
end

-- ---------------------
-- Events
-- ---------------------

function mod:COMBAT_LOG_EVENT_UNFILTERED(_, _, eventType, srcGUID, srcName, _, dstGUID, dstName, _, spellID, spellName, ...)
	if spellID ~= 50720 or srcName ~= playerName then return end
	if eventType == "SPELL_AURA_APPLIED" then
		self:StartBar(dstName, 30 * 60)
	elseif eventType == "SPELL_AURA_REFRESH" then
		--self:CancelBar()
		self:StartBar(dstName, 30 * 60)
	elseif eventType == "SPELL_AURA_REMOVED" then
		self:CancelBar()
	end
end

do
	local function GetLSMIndex(t, value)
		for k, v in pairs(Media:List(t)) do
			if v == value then
				return k
			end
		end
		return nil
	end

	local function set(t, value)
		db.profile.vigilance[t[#t]] = value
		self:ApplySettings()
	end
	local function get(t)
		return db.profile.vigilance[t[#t]]
	end
	
	local function setcolor(t, ...)
		db.profile.vigilance[t[#t]] = {...}
		self:ApplySettings()
	end
	local function getcolor(t)
		return unpack(db.profile.vigilance[t[#t]])
	end
	
	local function dragstart()
		vgParent:StartMoving()
	end
	local function dragstop()
		db.profile.vigilance.x = vgParent:GetLeft()
		db.profile.vigilance.y = vgParent:GetBottom()
		vgParent:StopMovingOrSizing()
	end
	
	local function testbar()
		self:TestBar()
	end
	
	-- Select tables
	local positions = { L["Left"], L["Center"], L["Right"] }
	
	Vigilant.options.args.Vigilance = {
		type = "group",
		name = L["Vigilance"],
		desc = L["Vigilance_Desc"],
		cmdHidden = true,
		disabled = function() return not self:IsEnabled() end,
		args = {
			desc = {
				type = "description",
				name = "  " .. L["Vigilance_Desc"] .. "\n\n",
				order = 1,
				cmdHidden = true,
				image = "Interface\\Icons\\ability_warrior_vigilance",
				imageWidth = 32, imageHeight = 32,
			},
			test = {
				type = "execute",
				name = "Test - DEBUGGING",
				desc = "Test the bar without actually having Vigilance up.",
				func = testbar,
				order = 4,
				hidden = true,
			},
			
			frame = {
				type = "group",
				name = L["Frame"],
				desc = nil,
				order = 100,
				inline = true,
				args = {
					lock = {
						type = "toggle",
						name = L["Lock"],
						desc = L["Toggle bar lock"],
						get = function(info)
							return self.locked
						end,
						set = function(info, v)
							self.locked = v
							if not self:IsEnabled() then return end
							if v then
								vgParent:EnableMouse(false)
								vgParent:SetScript("OnDragStart", nil)
								vgParent:SetScript("OnDragStop", nil)
							else
								vgParent:Show()
								vgParent:EnableMouse(true)
								vgParent:SetScript("OnDragStart", dragstart)
								vgParent:SetScript("OnDragStop", dragstop)
								self:ApplySettings()
							end
						end,
						order = 101,
					},
					blank1 = {
						type = "description",
						name = "",
						order = 102,
						cmdHidden = true,
						width = "full",
					},
					width = {
						type = "range",
						name = L["Width"],
						get = get, set = set,
						min = 10, max = 600, step = 1, bigStep = 5,
						order = 103,
					},
					height = {
						type = "range",
						name = L["Height"],
						get = get, set = set,
						min = 2, max = 50, step = 1,
						order = 104,
					},
					blank2 = {
						type = "description",
						name = "",
						order = 105,
						cmdHidden = true,
						width = "full",
					},
					scale = {
						type = "range",
						name = L["Scale"],
						get = get, set = set,
						min = 1, max = 150, step = 1, bigStep = 1,
						order = 106,
					},
					alpha = {
						type = "range",
						name = L["Alpha"],
						get = get, set = set,
						min = 0, max = 1, step = 0.1,
						order = 107,
						--width = "full",
					},
					blank3 = {
						type = "description",
						name = "",
						order = 108,
						cmdHidden = true,
						width = "full",
					},
					posX = {
						type = "input",
						name = L["X Position"],
						get = function(info) return tostring(db.profile.vigilance.x) end,
						set = function(info, v)
							db.profile.vigilance.x = tonumber(v)
							self:ApplySettings()
						end,
						order = 109,
					},
					posY = {
						type = "input",
						name = L["Y Position"],
						get = function(info) return tostring(db.profile.vigilance.y) end,
						set = function(info, v)
							db.profile.vigilance.y = tonumber(v)
							self:ApplySettings()
						end,
						order = 110,
					},
					blank4 = {
						type = "description",
						name = "",
						order = 111,
						cmdHidden = true,
						width = "full",
					},
					backColor = {
						type = "color",
						name = L["Background color"],
						get = getcolor, set = setcolor,
						hasAlpha = true,
						order = 112,
						--width = "full",
					},
					border = {
						type = "select",
						name = L["Border"],
						get = function(info) return GetLSMIndex("border", db.profile.vigilance.border) end,
						set = function(info, v)
							db.profile.vigilance.border = Media:List("border")[v]
							self:ApplySettings()
						end,
						values = Media:List("border"),
						order = 113,
						--width = nil,
					},
					blank5 = {
						type = "description",
						name = "",
						order = 114,
						cmdHidden = true,
						width = "full",
					},
					icon = {
						type = "toggle",
						name = L["Show icon"],
						get = function(info) return db.profile.vigilance.icon end,
						set = function(info, v)
							db.profile.vigilance.icon = v
							self:ApplySettings()
						end,
						order = 115,
					},
					iconPosition = {
						type = "select",
						name = L["Icon position"],
						get = get,
						set = function(info, v)
							db.profile.vigilance.iconPosition = v
							self:ApplySettings()
						end,
						values = { positions[1], positions[3] },
						disabled = function() return not db.profile.vigilance.icon end,
						order = 116,
					},
				}
			},
			
			bars = {
				type = "group",
				name = L["Bars"],
				order = 200,
				inline = true,
				args = {
					barColorByClass = {
						type = "toggle",
						name = "Color bar by class",
						desc = L["Use the class's color for the bar color"],
						get = function(info) return db.profile.vigilance.barColorByClass end,
						set = function(info, v)
							db.profile.vigilance.barColorByClass = v
							self:ApplySettings()
						end,
						order = 200,
					},
					blank1 = {
						type = "description",
						name = "",
						order = 201,
						cmdHidden = true,
						width = "full",
					},
					barColor = {
						type = "color",
						name = L["Bar color"],
						desc = L["Color of the bar"],
						get = getcolor, set = setcolor,
						hasAlpha = true,
						disabled = function() return db.profile.vigilance.barColorByClass end,
						order = 202,
					},
					texture = {
						type = "select",
						name = L["Texture"],
						get = function(info) return GetLSMIndex("statusbar", db.profile.vigilance.texture) end,
						set = function(info, v)
							db.profile.vigilance.texture = Media:List("statusbar")[v]
							self:ApplySettings()
						end,
						values = Media:List("statusbar"),
						disabled = function() return db.profile.vigilance.barColorByClass end,
						order = 205,
					},
				}
			},
			
			text = {
				type = "group",
				name = L["Text"],
				order = 300,
				inline = true,
				args = {
					textShow = {
						type = "toggle",
						name = L["Show text"],
						desc = L["Show countdown text."],
						get = get, set = set,
						order = 301,
					},
					blank1 = {
						type = "description",
						name = "",
						order = 302,
						cmdHidden = true,
						width = "full",
					},
					textColor = {
						type = "color",
						name = L["Text color"],
						desc = L["Text color"],
						get = getcolor,	set = setcolor,
						hasAlpha = true,
						order = 303,
					},
					textPosition = {
						type = "select",
						name = L["Text position"],
						desc = L["Text position"],
						get = get,
						set = function(info, v)
							db.profile.vigilance.textPosition = v
							self:ApplySettings()
						end,
						values = positions,
						order = 304,
					},
					blank3 = {
						type = "description",
						name = "",
						order = 305,
						cmdHidden = true,
						width = "full",
					},
					textSize = {
						type = "range",
						name = L["Text size"],
						desc = L["Text size"],
						get = get, set = set,
						min = 6, max = 20, step = 1,
						order = 306,
					},
					textFont = {
						type = "select",
						name = L["Text font"],
						desc = L["Text font"],
						get = function(info) return GetLSMIndex("font", db.profile.vigilance.textFont) end,
						set = function(info, v)
							db.profile.vigilance.textFont = Media:List("font")[v]
							self:ApplySettings()
						end,
						values = Media:List("font"),
						order = 307,
					},
				}
			},
		},
	}
	Vigilant.options.args.Vigilance.args.output = mod:GetSinkAce3OptionsDataTable()
end
